"""This module contains a function to run `pipeline` in a loop over multiple sources 
(written by Jeff Jennings)."""

import os
import json

def main(source_par_f='./pars_source.json', gen_par_f='./pars_gen.json'):
    disk_names = []
    source_pars = json.load(open(source_par_f, 'r'))
    for ii in source_pars:
        disk_names.append(ii)

    # run the radial pipeline over each source in `disk_names`
    for ii, jj in enumerate(disk_names):
        print("\nPipeline call {} of {} - disk {}".format(ii, len(disk_names) - 1, jj))

        os.system('python -m arksia.pipeline -d {} -b {} -s {}'.format(
            jj, gen_par_f, source_par_f))

if __name__ == "__main__":
    main()



"""This module contains a function for generating images from visibilities 
(written by Jeff Jennings)."""

import numpy as np 

from mpol.gridding import DirtyImager

def dirty_image(vis, robust, npix, pixel_scale, casa_vis=True):
    """
    Produce a dirty image (2D array) given visibilities

    Parameters
    ----------
    vis : list
        visibilities: u-coordinates, v-coordinates, visibility amplitudes 
        (Re(V) + Im(V) * 1j), weights   
    robust : float
        Robust weighting parameter.
    npix : int, default=None
        Number of pixels in image. 
    pixel_scale : float, default=None
        Pixel width [arcsec]. 
    casa_vis : bool, default=True
        Whether the 'vis' were produced with CASA, in which case their 
        complex conjugate will be taken to image with MPoL
        
    Returns
    -------
    im : array
        Dirty image
    """    
    if casa_vis:
        # MPoL uses the standard baseline convention, CASA doesn't
        vis[2] = np.conj(vis[2])

    # generate dirty image at same pixel scale as clean image
    imager = DirtyImager.from_image_properties( 
        cell_size=pixel_scale,
        npix=npix,
        uu=vis[0] / 1e3,
        vv=vis[1] / 1e3,
        weight=vis[3],
        data_re=vis[2].real,
        data_im=vis[2].imag,
        )
    im, _ = imager.get_dirty_image(weighting="briggs", robust=robust, unit='Jy/arcsec^2')

    return np.squeeze(im)


"""This module is the main file for running radial pipeline fits and analysis 
(written by Jeff Jennings)."""

import os; os.environ.get('OMP_NUM_THREADS', '1')
import json
import argparse
import numpy as np
import matplotlib.pyplot as plt
import multiprocess
import pickle 

from frank.constants import deg_to_rad
from frank.utilities import jy_convert
from frank.debris_fitters import FrankDebrisFitter
from frank.io import save_fit
from frank.make_figs import make_quick_fig
from frank.geometry import FixedGeometry
from frank.utilities import get_fit_stat_uncer

import arksia 
from arksia import input_output, extract_radial_profile, plot, parametric_fitter

def parse_parameters(*args):
    """
    Read in a .json parameter files to run the pipeline

    Parameters
    ----------
    parameter_filename : str
        Parameter file (.json)

    Returns
    -------
    args : dict
        Dictionary containing model parameters the pipeline uses
    """

    parser = argparse.ArgumentParser("Run the radial profile pipeline for ARKS data")

    parser.add_argument("-d", "--disk",
                        type=str,
                        help="Disk name")

    parser.add_argument("-b", "--base_parameter_filename",
                        type=str,
                        default="./pars_gen.json",
                        help="Parameter file (.json) with generic pars")
    
    parser.add_argument("-s", "--source_parameter_filename",
                        type=str,
                        default="./pars_source.json",
                        help="Parameter file (.json) with source-specific pars")

    args = parser.parse_args(*args)
    
    return args


def model_setup(parsed_args):
    """
    Initialze a set of model parameters to be used in pipeline routines.
    Run checks to catch errors on values of some parameters. 

    Parameters
    ----------
    parsed_args : dict
        Dictionary containing model parameters to be used in the pipeline

    Returns
    -------
    model : dict
        Dictionary containing final model parameters the pipeline uses, with 
        some processing and added parameters relative to `parsed_args`
    """

    # generic parameters
    model = json.load(open(parsed_args.base_parameter_filename, 'r'))
    model["base"]["disk"] = parsed_args.disk

    print('\nRunning radial profile pipeline for {}'.format(model["base"]["disk"]))

    # disk-specific parameters
    source_pars = json.load(open(parsed_args.source_parameter_filename, 'r'))
    disk_pars = source_pars[model["base"]["disk"]]

    # expect input files to be in "<root_dir>/<disk name>/<clean or rave or frank>"
    model["base"]["save_dir"] = os.path.join(model["base"]["root_dir"], "{}".format(model["base"]["disk"]))
    print("  Model setup: setting load/save paths as {}/<clean, rave, frank>. Visibility tables should be in frank path.".format(model["base"]["save_dir"]))

    model["base"]["clean_dir"] = os.path.join(model["base"]["save_dir"], "clean")
    model["base"]["rave_dir"] = os.path.join(model["base"]["save_dir"], "rave")
    model["base"]["frank_dir"] = os.path.join(model["base"]["save_dir"], "frank")
    model["base"]["parametric_dir"] = os.path.join(model["base"]["save_dir"], "parametric")
    
    subdirs = model["base"]["clean_dir"], model["base"]["rave_dir"], \
        model["base"]["frank_dir"], model["base"]["parametric_dir"]
    for dd in subdirs: 
        os.makedirs(dd, exist_ok=True)

    if disk_pars["base"]["SMG_sub"] is True:
        model["base"]["SMG_sub"] = "SMGsub."
    else:
        model["base"]["SMG_sub"] = ""

    model["base"]["dist"] = disk_pars["base"]["dist"]

    model["clean"]["npix"] = disk_pars["clean"]["npix"]
    model["clean"]["pixel_scale"] = disk_pars["clean"]["pixel_scale"]

    # get clean image rms
    image_pars = json.load(open(os.path.join(model["base"]["root_dir"], "pars_image.json"), 'r'))
    disk_image_pars = image_pars[model["base"]["disk"]]
    robusts, rmss = disk_image_pars["clean"]["image_robust"], disk_image_pars["clean"]["image_rms"]
    model["clean"]["image_rms"] = rmss[robusts.index(model["clean"]["robust"])]

    model["clean"]["bestfit"] = {}
    model["clean"]["bestfit"]["robust"] = disk_pars["clean"]["bestfit"]["robust"]

    model["rave"]["pixel_scale"] = disk_pars["rave"]["pixel_scale"]
    
    model["frank"]["bestfit"] = {}
    model["frank"]["bestfit"]["alpha"] = disk_pars["frank"]["bestfit"]["alpha"]
    model["frank"]["bestfit"]["wsmooth"] = disk_pars["frank"]["bestfit"]["wsmooth"]
    model["frank"]["bestfit"]["method"] = disk_pars["frank"]["bestfit"]["method"]

    # get source geom for clean profile extraction and frank fit
    mcmc = json.load(open(os.path.join(model["base"]["save_dir"], "MCMC_results.json"), 'r'))
    try: 
        dRA = mcmc["deltaRA-12mLB.obs1"]["median"]
        dDec = mcmc["deltaDec-12mLB.obs1"]["median"]
    except KeyError:
        dRA = mcmc["deltaRA-12m.obs1"]["median"]
        dDec = mcmc["deltaDec-12m.obs1"]["median"]

    geom = {"inc" : mcmc["i"]["median"],
            "PA" : mcmc["PA"]["median"], 
            "dRA" : dRA,
            "dDec" : dDec
            }
    model["base"]["geom"] = geom 
    print('    source geometry from MCMC {}'.format(geom))

    # stellar flux to remove from visibilities as point-source
    if model["frank"]["set_fstar"] == "custom":
        model["frank"]["fstar"] = disk_pars["frank"]["custom_fstar"] / 1e6
    elif model["frank"]["set_fstar"] == "SED":
        model["frank"]["fstar"] = disk_pars["frank"]["SED_fstar"] / 1e6
    elif model["frank"]["set_fstar"] == "MCMC":
        try:
            model["frank"]["fstar"] = mcmc["fstar"]["median"] / 1e3
        except KeyError:
            model["frank"]["fstar"] = 0.0
            print('        no stellar flux in MCMC file --> setting fstar = 0')
    else:
        raise ValueError("Parameter ['frank']['set_fstar'] is '{}'. It should be one of 'MCMC', 'SED', 'custom'".format(model["frank"]["set_fstar"])) 

    # enforce a Normal fit if finding scale height (LogNormal fit not compatible with vertical inference)
    if model["base"]["run_frank"] is True and model["frank"]["scale_heights"] is not None:
        print("    'scale_heights' is not None in your parameter file -- enforcing frank 'method=Normal' and 'max_iter=2000'")
        model["frank"]["method"] = "Normal"
        model["frank"]["max_iter"] = 2000

    # implemented functional forms 
    valid_funcs = [x for x in dir(arksia.parametric_forms) if not x.startswith('__')]
    if model["base"]["run_parametric"] is True:
        for pp in model["parametric"]["form"]:
            if pp not in valid_funcs:
                raise ValueError(f"{pp} is not one of {valid_funcs}")

    return model


def extract_clean_profile(model):
    """Obtain radial profiles from each of a CLEAN image and CLEAN model 

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters

    Returns
    -------
    tuple
        Radial points `r` [arcsec], brightness `I` [Jy/sr] and brightness 
        uncertainty `I_err` [Jy/sr] for each the CLEAN image profile' `r` and 
        `I` for the CLEAN model profile
    """
    # image filenames 
    base_path = "{}/{}.combined.{}corrected.briggs.{}.{}.{}".format(
        model["base"]["clean_dir"], 
        model["base"]["disk"], 
        model["base"]["SMG_sub"],
        model["clean"]["robust"], 
        model["clean"]["npix"], 
        model["clean"]["pixel_scale"]
        )

    # get image arrays
    clean_fits = base_path + ".pbcor.fits"
    pb_fits = base_path + ".pb.fits"
    model_fits = base_path + ".model.fits"

    clean_image, clean_beam = input_output.load_fits_image(clean_fits)
    bmaj, bmin = clean_beam
    pb_image = input_output.load_fits_image(pb_fits, aux_image=True)
    model_image = input_output.load_fits_image(model_fits, aux_image=True)

    # profile of clean image.
    # for radial profile on east side of disk,
    # range in azimuth (PA +- range) over which to average 
    f = 1.3
    phic_rad = extract_radial_profile.find_phic(model["base"]["geom"]["inc"] * np.pi / 180, f)
    phic_deg = phic_rad / deg_to_rad
    

    print('  Clean profiles: extracting profiles from {} and {} using phi_crit {:.2f} deg'.format(clean_fits, model_fits, phic_deg))

    phis_E = np.linspace(model["base"]["geom"]["PA"] - phic_deg, 
                        model["base"]["geom"]["PA"] + phic_deg, 
                        model["clean"]["Nphi"]
                        ) 
    
    phis_W = phis_E + 180

    # radial profile of east and west sides
    r_E, I_E, I_err_E = extract_radial_profile.radial_profile_from_image(
        clean_image, geom=model["base"]["geom"], phis=phis_E, bmaj=bmaj, 
        bmin=bmin, pb_image=pb_image, **model["clean"])
    r_W, I_W, I_err_W = extract_radial_profile.radial_profile_from_image(
        clean_image, geom=model["base"]["geom"], phis=phis_W, bmaj=bmaj,
        bmin=bmin, pb_image=pb_image, **model["clean"])

    # average of E and W
    r, I, I_err = r_W, np.mean((I_E, I_W), axis=0), np.hypot(I_err_E, I_err_W) / 2


    # profile of CLEAN .model image.
    # average across all azimuths (no need to take separate E and W profiles)
    phis_mod = np.linspace(model["base"]["geom"]["PA"] - 180, 
                                  model["base"]["geom"]["PA"] + 180,
                                  model["clean"]["Nphi"] 
                                  )
    
    r_mod, I_mod = extract_radial_profile.radial_profile_from_image(
        model_image, geom=model["base"]["geom"], phis=phis_mod, bmaj=0, 
        bmin=0, model_image=True, **model["clean"])

    # radial profile save paths
    ciff = "{}/clean_profile_robust{}.txt".format(
        model["base"]["clean_dir"], model["clean"]["robust"])
    cmff = "{}/clean_model_profile_robust{}.txt".format(
        model["base"]["clean_dir"], model["clean"]["robust"])    
    
    print('    saving CLEAN image profile to {} and model profile to {}'.format(ciff, cmff))
    np.savetxt(ciff, 
        np.array([r, I, I_err]).T, 
        header='Extracted from {}\nr [arcsec]\tI [Jy/sr]\tI_err [Jy/sr]'.format(
            clean_fits.split('/')[-1])
        )
    np.savetxt(cmff,
        np.array([r_mod, I_mod]).T,        
        header='Extracted from {}\nr [arcsec]\tI [Jy/sr]'.format(
            model_fits.split('/')[-1])
        )
 

def process_rave_fit(model):
    """Unpack a fitted RAVE radial profile, convert units, save

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters

    Returns
    -------
    list
        Radial points `r` [arcsec], brightness `I` [Jy/sr] and brightness 
        uncertainty `I_err` [Jy/sr] for the RAVE radial profile
    """
    
    fit_path = input_output.parse_rave_filename(model, file_type='rave_fit')
    print('  Rave profiles: processing {}'.format(fit_path))

    r, I_err_lo, I, I_err_hi = np.load(fit_path)
    I_err_lo = I - I_err_lo
    I_err_hi = I_err_hi - I

    I = jy_convert(I, 'arcsec2_sterad')
    I_err_lo = jy_convert(I_err_lo, 'arcsec2_sterad')
    I_err_hi = jy_convert(I_err_hi, 'arcsec2_sterad')    

    ff = "{}/rave_profile_robust{}.txt".format(
        model["base"]["rave_dir"], model["clean"]["robust"])
    print('    saving RAVE profile to {}'.format(ff))

    np.savetxt(ff, 
        np.array([r, I, I_err_lo, I_err_hi]).T, 
        header='Extracted from {}\nr [arcsec]\tI [Jy/sr]\tI_err (lower bound) [Jy/sr]\tI_err (upper bound) [Jy/sr]'.format(
            fit_path.split('/')[-1])
        )


def run_frank(model):
    """Perform a frank fit

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters

    Returns
    -------
    sols : list of _HankelRegressor objects
        frank fit solutions for each set of hyperparameters 
        (see frank.radial_fitters.FrankFitter)
    """
    print(' Frank fit: running fit')

    uv_data = input_output.get_vis(model)

    print('    shifting visibilities down by fstar {} uJy according to {}'.format(model["frank"]["fstar"] * 1e6, model["frank"]["set_fstar"]))
    uv_data[2] = uv_data[2] - model["frank"]["fstar"]
    
    frank_geom = FixedGeometry(**model["base"]["geom"]) 

    # set scale height
    if model["frank"]["scale_heights"] is None:
        hs = [0]
    else:
        hs = np.logspace(*model["frank"]["scale_heights"])
        print("    aspect ratios to be sampled: {}".format(hs))      

    # perform frank fit(s)
    def frank_fitter(priors):
        alpha, wsmooth, h = priors 
        print('        performing fit for alpha {} wsmooth {} h {}'.format(alpha, wsmooth, h))

        if h == 0:
            scale_height = None  
        else:
            def scale_height(R):
                return h * R
            
        # FrankDebrisFitter assumes disk is optically thin
        FF = FrankDebrisFitter(Rmax=model["frank"]["rout"], 
                                N=model["frank"]["N"], 
                                geometry=frank_geom,
                                scale_height=scale_height, 
                                alpha=alpha,
                                weights_smooth=wsmooth,
                                method=model["frank"]["method"],
                                I_scale=model["frank"]["I_scale"],
                                max_iter=model["frank"]["max_iter"],
                                convergence_failure='warn'
                                )
        
        sol = FF.fit(*uv_data)

        # add non-negative brightness profile to sol
        if model["frank"]["method"] == "Normal":
            setattr(sol, '_nonneg', sol.solve_non_negative())

        if scale_height is None:
            logev = None
        else:
            logev = FF.log_evidence_laplace()
        # add evidence to sol
        setattr(sol, 'logevidence', logev)

        # save fit outputs
        save_prefix = "{}/{}_alpha{}_w{}_h{:.3f}_fstar{:.0f}uJy_method{}".format(
                    model["base"]["frank_dir"], model["base"]["disk"], 
                    alpha, wsmooth, h,
                    model["frank"]["fstar"] * 1e6,
                    model["frank"]["method"]
                    )

        print("          saving fit results to {}*".format(save_prefix))
        save_fit(*uv_data, 
                sol=sol,
                prefix=save_prefix,
                save_solution=model["frank"]["save_solution"],
                save_profile_fit=model["frank"]["save_profile_fit"],
                save_vis_fit=model["frank"]["save_vis_fit"],
                save_uvtables=model["frank"]["save_uvtables"]
                )

        # save fit summary figures
        if model["frank"]["make_quick_fig"]:
            qfig, qaxes = make_quick_fig(*uv_data, sol, bin_widths=model["plot"]["bin_widths"],
                        save_prefix=None,
                        )
            # add 1 sigma statistical uncertainty band to plot
            sigmaI = get_fit_stat_uncer(sol)
            qaxes[0].fill_between(sol.r, (sol.I - sigmaI) / 1e10, (sol.I + sigmaI) / 1e10, color='r', alpha=0.2)
            qaxes[1].fill_between(sol.r, (sol.I - sigmaI) / 1e10, (sol.I + sigmaI) / 1e10, color='r', alpha=0.2)
            plt.savefig(save_prefix + '_frank_fit_quick.png', dpi=300,
                        bbox_inches='tight')
            plt.close()            

            # reprojected frank residual visibilities
            frank_resid_vis = [uv_data[0], uv_data[1], uv_data[2] - sol.predict(uv_data[0], uv_data[1]), uv_data[3]]
            plot.frank_image_diag_figure(model, sol, frank_resid_vis, save_prefix=save_prefix)

        return sol

    nfits = len((model["frank"]["alpha"])) * len(model["frank"]["wsmooth"]) * len(hs)
    if nfits == 1:
        # import frank; frank.enable_logging()
        sol = frank_fitter([model["frank"]["alpha"][0], model["frank"]["wsmooth"][0], hs[0]])
        return sol
    
    else: 
        # use as many threads as there are fits, up to a maximum of 'model["frank"]["nthreads"]'
        nthreads = min(nfits, model["frank"]["nthreads"])
        pool = multiprocess.Pool(processes=nthreads)
        print('    {} fits will be performed. Using {} threads (1 thread per fit; {} threads available on CPU).'.format(nfits, 
                                                                                             nthreads, 
                                                                                             multiprocess.cpu_count())
                                                                                             )

        # grids of prior values
        g0, g1, g2 = np.meshgrid(model["frank"]["alpha"], model["frank"]["wsmooth"], hs)
        g0, g1, g2 = g0.flatten(), g1.flatten(), g2.flatten()
        priors = np.array([g0, g1, g2]).T
        # run fits over grids
        sols = pool.map(frank_fitter, priors)

        logevs = []
        for ss in sols:
            logevs.append(ss.logevidence)

        # save scale heights and log evidences
        if np.array(hs).any() != 0:
            ff = "{}/vertical_inference_frank.txt".format(model["base"]["save_dir"])
            print("    saving h and log evidence results to {}".format(ff))
            np.savetxt(ff,
                np.array([g0, g1, g2, logevs]).T, header='alpha\twsmooth\th=H/r\tlog evidence'
            )

        return sols
    

def fit_parametric(fits, model):
    """
    # TODO
    """

    # get frank best-fit profile
    _, _, results = fits
    frank_profile = results[0]

    # run parametric fits
    figs = []
    PFits = []
    for pp in model['parametric']['form']:
        PFit = parametric_fitter.ParametricFit(frank_profile, 
                                            model, 
                                            func_form=pp,
                                            learn_rate=model['parametric']['learn_rate'], 
                                            niter=int(model['parametric']['niter']))
        PFit.fit()
        PFits.append(PFit)

        print(f"    initial params {PFit.initial_params}\n    final {PFit.bestfit_params}\n    loss {PFit.loss_history}")

        ff = f"{model['base']['parametric_dir']}/parametric_fit_{pp}.obj"
        print(f"    saving parametric fit results to {ff}")
        with open(ff, 'wb') as f:
            pickle.dump(PFit, f)

        # plot results
        fig = plot.parametric_fit_figure(PFit, frank_profile, model)
        figs.append(fig)

    return PFits, frank_profile, figs


def main(*args):
    """Run a pipeline to extract/fit/analyze radial profiles using 
    clean/rave/frank for ARKS data.

    Parameters
    ----------
    *args : strings
        Simulates the command line arguments
    """

    parsed_args = parse_parameters(*args)
    model = model_setup(parsed_args)
    
    if model["base"]["extract_clean_profile"] is True:
        extract_clean_profile(model)

    if model["base"]["process_rave_fit"] is True:
        process_rave_fit(model)

    if model["base"]["run_frank"] is True:
        frank_sols = run_frank(model)

    if model["base"]["compare_models_fig"] is True:
        fits = input_output.load_bestfit_profiles(model)   
        fig1 = plot.profile_comparison_figure(fits, model)
        fig2 = plot.image_comparison_figure(fits, model)

    if model["base"]["aspect_ratio_fig"] is True:
        fig3 = plot.aspect_ratio_figure(model)

    if model["base"]["run_parametric"] is True:
        fits = input_output.load_bestfit_profiles(model)
        parametric_fits, frank_profile, figs = fit_parametric(fits, model)

if __name__ == "__main__":
    main()



"""This module contains a function to obtain results from fits/analysis of the full ARKS survey 
(written by Jeff Jennings)."""

import json
import numpy as np 
import matplotlib.pyplot as plt 

from arksia.pipeline import model_setup
from arksia.input_output import load_bestfit_profiles

def main(source_par_f='./pars_source.json', gen_par_f='./pars_gen.json', 
         profiles_txt=True, profiles_fig=True, robust=2.0
                   ):
    """
    Generate radial profile results across all survey sources.

    Parameters
    ----------
    source_par_f : string, default='pars_source.json'
        Path to the parameter file with custom values for each source        
    gen_par_f : string, default='pars_gen.json'
        Path to the general parameter file    
    profiles_txt : bool, default=True
        Whether to produce a .txt file per source containting the 
        clean, rave, frank brightness profiles (sampled at same radii)
    profiles_fig : bool, default=True
        Whether to produce a single figure showing brightness profiles for 
        all sources
    robust : float, default=2.0
        Robust weighting value to use for retrieving clean, rave results
    
    Returns
    -------
    figs : `plt.figure` instance
        The generated figures, produced if `profiles_fig` is True
    """

    # get all source names
    source_pars = json.load(open(source_par_f, 'r'))
    disk_names = []
    for dd in source_pars:
        disk_names.append(dd)

    fig0, axs0 = plt.subplots(nrows=5, ncols=4, figsize=(10, 10), squeeze=True)
    fig1, axs1 = plt.subplots(nrows=5, ncols=4, figsize=(10, 10), squeeze=True)
    figs, axs = [fig0, fig1], [axs0, axs1]

    for ii, jj in enumerate(disk_names):
        # generate model for each source
        class parsed_args():
            base_parameter_filename = gen_par_f
            source_parameter_filename = source_par_f
            disk = jj                        
        model = model_setup(parsed_args)

        # best-fit clean, rave, frank profile for each source
        [[rc, Ic, Iec], [grid, Vc]], [[rr, Ir, Ier_lo, Ier_hi], [grid, Vr]], [[rf, If, Ief], [grid, Vf], sol] = load_bestfit_profiles(model, robust)

        # interpolate clean and rave profiles onto frank radial points 
        Ic_interp = np.interp(rf, rc, Ic)
        Iec_interp = np.interp(rf, rc, Iec)
        Ir_interp = np.interp(rf, rr, Ir)
        Ier_lo_interp = np.interp(rf, rr, Ier_lo)
        Ier_hi_interp = np.interp(rf, rr, Ier_hi)

        Is_interp = [Ic_interp, Ir_interp, If]
        Ies_interp = [[Iec_interp, Iec_interp], [Ier_lo_interp, Ier_hi_interp], [Ief, Ief]]


        if profiles_txt:
            ff = '{}/{}_radial_profiles.txt'.format(model["base"]["save_dir"], jj)
            print('  Survey summary: saving radial profiles to {}'.format(ff))

            # save .txt file per source with clean,rave,frank profiles
            np.savetxt(ff, 
                    np.array([rf * model["base"]["dist"], Ic_interp, Iec_interp, 
                              If, Ief, Ir_interp, Ier_lo_interp, Ier_hi_interp
                              ]).T,
                    header="dist={} [au].\nAll brightnesses in [Jy/steradian].\nUncertainties not comparable across models. " + \
                    "Rave uncertainties have unique lower and upper bounds.\nColumns: " + \
                    "r [au]\t\tI_clean\t\tsigma_clean\t\tI_frank\t\tsigma_frank\t\tI_rave\t\tsigma_lower_rave\t\tsigma_upper_rave".format(model["base"]["dist"])
                    )


        if profiles_fig:
            # generate, save figures for brightness profiles of all sources and brightness profiles with uncertainties
            for hh in range(2):
                fig = figs[hh]
                ax = axs[hh]

                # flatten axes
                ax = [bb for aa in ax for bb in aa]
                cols, labs = ['C1', 'C3', 'C2'], ['clean', 'rave', 'frank']

                for kk, ll in enumerate(Is_interp):     
                    # plot profile
                    ax[ii].plot(rf * model["base"]["dist"], ll / 1e6, c=cols[kk], label=labs[kk])
                
                    if hh == 1:
                        # 1 sigma uncertainty band
                        ax[ii].fill_between(rf * model["base"]["dist"], 
                                            (ll - Ies_interp[kk][0]) / 1e6, (ll + Ies_interp[kk][1]) / 1e6, 
                                        color=cols[kk], alpha=0.4)
                
                ax[ii].axhline(y=0, ls='--', c='k')

                ax[ii].text(0.6, 0.9, jj, transform=ax[ii].transAxes)

                if ii == len(disk_names) - 1:
                    ax[ii].legend(loc='center right')
                    ax[ii].set_xlabel('r [au]')
                    ax[ii].set_ylabel(r'I [$10^5$ Jy sterad$^{-1}$]')

                ax[ii].set_title('$f_*=${:.0f} $\mu$Jy'.format(model["frank"]["fstar"] * 1e6))

    if profiles_fig:
        print('  Survey summary: making survey summary figure')
    fig1.suptitle(r'$1\sigma$ uncertainties do not include systematic unc., and are not comparable across models')    

    ff0 = '{}/survey_profile_summary.png'.format(model["base"]["root_dir"])
    ff1 = '{}/survey_profile_summary_unc.png'.format(model["base"]["root_dir"])
    print('    saving figures to {} and {}'.format(ff0, ff1))

    plt.figure(fig0); plt.tight_layout(); plt.savefig(ff0, dpi=300)
    plt.figure(fig1); plt.tight_layout(); plt.savefig(ff1, dpi=300)

    return figs
    
if __name__ == "__main__":
    main()


"""This module contains functions for loading and saving products in the pipeline 
(written by Jeff Jennings)."""

import os
import numpy as np

import astropy.io.fits as pyfits

from frank.utilities import generic_dht, get_fit_stat_uncer
from frank.io import load_sol, load_uvtable

def concatenate_vis(in_paths, out_path):
    """
    Concatenate the visibilities from multiple files; save as one output .npz file.

    Parameters
    ----------
    in_paths : list of string
        Filenames of the input visibility tables
    out_path : string
        Filename of the output table

    Returns
    -------
    uv_data : list
        concatenated dataset: u-coordinates, v-coordinates, visibility amplitudes 
        (Re(V) + Im(V) * 1j), weights
    """
    u, v, re, im, weights = [], [], [], [], []
    for ii in in_paths:
        try: 
            ut, vt, ret, imt, weightst, _ = np.genfromtxt(ii).T
            u.extend(ut)
            v.extend(vt)
            re.extend(ret)
            im.extend(imt)
            weights.extend(weightst)
        except OSError:
            pass

    u, v, weights = np.array(u), np.array(v), np.array(weights)
    vis = np.add(re, np.multiply(im, 1j))

    np.savez(out_path,
             u=u, v=v, V=vis, weights=weights,
             units={'u': 'lambda', 'v': 'lambda',
                    'V': 'Jy', 'weights': 'Jy^-2'})

    return [u, v, vis, weights]


def get_vis(model):
    """
    Load (or generate if it does not exist) an ARKS visibility dataset.

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters

    Returns
    -------
    uv_data : list
        dataset: u-coordinates, v-coordinates, visibility amplitudes 
        (Re(V) + Im(V) * 1j), weights
    """

    combined_vis_path = "{}/vis_combined.npz".format(model["base"]["frank_dir"])

    if os.path.isfile(combined_vis_path):
        print('    loading combined visibility file {}'.format(combined_vis_path))
        dat = np.load(combined_vis_path)
        u, v, vis, weights = [dat[i] for i in ['u', 'v', 'V', 'weights']]
        uv_data = [u, v, vis, weights] 

    else:
        print('    combined visibility file {} not found. Creating it by combining .txt files.'.format(combined_vis_path))

        # input filenames
        vis_paths = []
        for ii in ['ACA', '12m', '12mSB', '12mLB']:
            vis_paths.append("{}/{}.{}.continuum.fav.tav.{}corrected.txt".format(
                model["base"]["frank_dir"], model["base"]["disk"], ii, model["base"]["SMG_sub"]
                )
            )

        uv_data = concatenate_vis(vis_paths, combined_vis_path)

    return uv_data 


def load_fits_image(fits_image, aux_image=False, verbose=0):
    """
    Load an image from a .fits file.

    Parameters
    ----------
    fits_image : string
        Path to the .fits file
    aux_image : bool, default=False
        Whether the .fits image is an 'auxillary' image (such as a dirty image) 
        or a standard clean output image
    verbose : int, default=0
        Level of verbosity for print statements
        
    Returns
    -------
    image, [bmaj, bmin] : 2D array, list of floats
        The image and its major and min axis widths (only `image` is returned 
        if `aux_image=True`)
    """

    ff = pyfits.open(fits_image)
    image = get_last2d(ff[0].data) 

    if verbose > 0:
        print('        loading .fits image {}, brightness unit {}'.format(fits_image, ff[0].header['BUNIT']))

    if not aux_image:
        header = ff[0].header
        bmaj = float(header['BMAJ'])
        bmin = float(header['BMIN'])
        # pixel_scale = float(header['CDELT2']) 
        # npix = len(im[:,0])

        return image, np.array([bmaj, bmin])

    return image 


def get_last2d(image):
    "Get last 2 dimensions of an N-dimensional image"
    if image.ndim <= 2:
        return image
    slc = [0] * (image.ndim - 2) + [slice(None), slice(None)]
    return image[tuple(slc)]


def parse_rave_filename(model, file_type='rave_fit'):
    """Interpret rave filenames for generalizing the loading of rave results.

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters
    file_type : str, default='rave_fit'
        Which type of rave file to load: 
          - 'rave_fit' for a rave brightness profile
          - 'rave_residual_image' for a 2d rave residual image

    Returns
    -------
    file_path : str
        Path to the desired rave file
    """    

    if model["clean"]["robust"] == 0.5:
        rave_str = "1"
    else:
        rave_str = "2"

    # one source has different rave parameter choices
    if model['base']['disk'] == "HD161868" and rave_str == "2":
        raveN = 7
    else: 
        raveN = 5

    if file_type == 'rave_fit':
        # adjust rave fit paths here if neeeded
        file_path = "{}/{}-{}_inc=90_N={}_radial_{}0arcsec.npy".format(
            model["base"]["rave_dir"], 
            model["base"]["disk"], 
            rave_str,
            raveN,
            model["rave"]["pixel_scale"]
            )
    
    elif file_type == 'rave_residual_image':
        file_path = "{}/{}-{}_inc=90_N={}_2Dresiduals.npy".format(
            model["base"]["rave_dir"], 
            model["base"]["disk"], 
            rave_str,
            raveN
            )        

    else: 
        raise ValueError("'file_type' {} must be one of 'rave_fit', 'rave_residual_image'".format(file_type))
    
    return file_path


def load_bestfit_frank_uvtable(model, resid_table=False):
    """
    Load the frank best-fit 2D reprojected visibilities.

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters
    resid_table : bool, default=False
        Whether the visibilities being loaded are the frank fit residuals 
        or the frank fit itself

    Returns
    -------
    uv_data : list
        frank visibilities: u-coordinates, v-coordinates, visibility amplitudes 
        (Re(V) + Im(V) * 1j), weights    
    """    
    if resid_table:
        suffix = 'resid'
    else:
        suffix = 'fit'

    # enforce the best-fit has 0 scale height
    path = "{}/{}_alpha{}_w{}_h0.000_fstar{:.0f}uJy_method{}_frank_uv_{}.npz".format(
                        model["base"]["frank_dir"], 
                        model["base"]["disk"], 
                        model["frank"]["bestfit"]["alpha"],
                        model["frank"]["bestfit"]["wsmooth"], 
                        model["frank"]["fstar"] * 1e6,
                        model["frank"]["bestfit"]["method"],
                        suffix
    )
    
    u, v, V, weights = load_uvtable(path)

    return [u, v, V, weights]


def load_bestfit_profiles(model):
    """
    Load the clean, rave and frank best-fit radial brightness profiles and 
    radial visibility profiles

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters

    Returns
    -------
    profiles : nested list
        - For clean: the brightness profile radial points, brightness values and 
        uncertainties; baselines and visibility amplitudes of the 1D Fourier 
        transform of the brightness profile.
        - For rave: the brightness profile radial points, brightness values and 
        unique lower and upper uncertainties; baselines and visibility 
        amplitudes of the 1D Fourier transform of the brightness profile.  
        - For frank: the brigthness profile radial points, brightness values and 
        uncertainties; baselines and visibility amplitudes of the visibility
        fit; the fit solution object.      
    """

    clean_bestfit = "{}/clean_profile_robust{}.txt".format(
        model["base"]["clean_dir"], model["clean"]["bestfit"]["robust"])
    rc, Ic, Iec = np.genfromtxt(clean_bestfit).T

    rave_bestfit = "{}/rave_profile_robust{}.txt".format(
        model["base"]["rave_dir"], model["clean"]["bestfit"]["robust"])
    rr, Ir, Ier_lo, Ier_hi = np.genfromtxt(rave_bestfit).T

    # enforce the best-fit has 0 scale height
    frank_bestfit = "{}/{}_alpha{}_w{}_h0.000_fstar{:.0f}uJy_method{}_frank_sol.obj".format(
                        model["base"]["frank_dir"], 
                        model["base"]["disk"], 
                        model["frank"]["bestfit"]["alpha"],
                        model["frank"]["bestfit"]["wsmooth"],
                        model["frank"]["fstar"] * 1e6,
                        model["frank"]["bestfit"]["method"],
    )
    sol = load_sol(frank_bestfit)
    rf, If, Ief = sol.r, sol.I, get_fit_stat_uncer(sol)

    # dense grid for visibility profiles
    grid = np.logspace(np.log10(1e3), np.log10(1e6), 10**3)

    # clean visibility profile
    _, Vc = generic_dht(rc, Ic, Rmax=sol.Rmax, N=sol._info["N"], grid=grid,
                            inc=0) # 'inc=0' passed to enforce optically thin assumption
    # rave visibility profile
    _, Vr = generic_dht(rr, Ir, Rmax=sol.Rmax, N=sol._info["N"], grid=grid,
                            inc=0)
    # frank visibility profile
    Vf = sol.predict_deprojected(grid)

    return [[rc, Ic, Iec], [grid, Vc]], \
        [[rr, Ir, Ier_lo, Ier_hi], [grid, Vr]], \
        [[rf, If, Ief], [grid, Vf], sol]


"""This module contains a class to fit parametric forms to a radial brightness signal 
(written by Jeff Jennings)."""

from tqdm.auto import tqdm
import jax.numpy as jnp
import jax
import optax

from arksia import parametric_forms

class ParametricFit():
    """
    # TODO
    """

    def __init__(self, truth, model, func_form, learn_rate=1e-3, niter=100000):
        # set jax device. Must come before any jax calls.
        if model["parametric"]["device"] is not None: # TODO
            self._device = model["parametric"]["device"]
            jax.config.update('jax_platform_name', self._device)
        else:
            self._device = jax.default_backend()
        print(f"    JAX is using the {self._device}.")

        # get frank brightness profile that we'll fit a parametric form to.
        # 'x' is radial points of frank fit, 'y' is brightness, 'err' is lower
        # bound on brightness uncertainty.
        x, y, err = truth
        self._x, self._y, self._err = jnp.array(x), jnp.array(y), jnp.array(err)

        self._model = model 

        self._learn_rate = learn_rate
        self._niter = niter

        self._form = func_form

        self._initial_params = {}


    def parametric_model(self, params: optax.Params, x: jnp.ndarray):
        """
        # TODO
        """
        
        form = self._form

        if form == 'asym_gauss':
            return parametric_forms.asym_gauss(params, x) 
        if form == 'triple_gauss':
            return parametric_forms.triple_gauss(params, x) 
        
        elif form == 'double_powerlaw':
            return parametric_forms.double_powerlaw_erf(params, x)
        elif form == 'double_powerlaw_erf':
            return parametric_forms.double_powerlaw_erf(params, x)
        elif form == 'double_powerlaw_gauss':
            return parametric_forms.double_powerlaw_gauss(params, x)
        elif form == 'double_powerlaw_double_gauss':
            return parametric_forms.double_powerlaw_double_gauss(params, x)
                
        elif form == 'single_erf_powerlaw':
            return parametric_forms.single_erf_powerlaw(params, x)
        elif form == 'double_erf_powerlaw':
            return parametric_forms.double_erf_powerlaw(params, x)
        
        else:
            raise ValueError(f"{form} invalid")
        

    def loss(self, params: optax.Params, x: jnp.ndarray, y: jnp.ndarray, 
             err: jnp.ndarray) -> jnp.ndarray:
        """
        # TODO
        """
        
        y_hat = self.parametric_model(params, x)

        # return (jnp.mean((y - y_hat) ** 2)) ** 0.5 # RMSE loss
        return ((y - y_hat) ** 2 / err ** 2).sum()



    def _fit(self, params: optax.Params, optimizer: optax.GradientTransformation, 
            niter: int) -> tuple[optax.Params, jnp.ndarray]:
        """
        # TODO
        """

        opt_state = optimizer.init(params)

        @jax.jit
        def step(params, opt_state, x, y, err):
            loss_value, grads = jax.value_and_grad(self.loss)(params, x, y, err)

            updates, opt_state = optimizer.update(grads, opt_state, params)
            params = optax.apply_updates(params, updates)

            return params, opt_state, loss_value

        loss_arr = jnp.zeros(niter)
        loss_initial = self.loss(params, self._x, self._y, self._err)
        loss_arr = loss_arr.at[0].set(loss_initial)

        progress = tqdm(range(1, niter))

        for i in progress:
            params, opt_state, loss_current = step(params, opt_state, 
                                                  self._x, self._y, self._err)
            loss_arr = loss_arr.at[i].set(loss_current)

            if i % 100 == 0:
                progress.set_description(f"    loss {loss_arr[i]:.2f}")

        self._loss_history = loss_arr
        self._params = params

        return


    def fit(self):
        """
        # TODO
        """

        form = self._form
        print(f"    fitting parametric form {form} to input profile")

        # set initial guesses for parameter values by using the signal we're fitting:
        # centroid or critical radius
        peak_idx = jnp.argmax(self._y)
        Rc = self._x[peak_idx]

        # amplitude
        amax = max(self._y)

        # standard deviation (obtain by guessing FWHM)
        idx = next(i for i,j in enumerate(self._y) if
                   i > peak_idx and j < amax / 2)
        fwhm = (self._x[idx] - Rc) * 2
        sigma = fwhm / (8 * jnp.log(2)) ** 0.5

        if form == 'asym_gauss':
            self._initial_params["Rc"] = Rc
            self._initial_params["a"] = amax
            self._initial_params["sigma1"] = sigma
            self._initial_params["sigma2"] = sigma

        elif form == 'triple_gauss':
            self._initial_params["R1"] = 0.7 * Rc
            self._initial_params["R2"] = Rc
            self._initial_params["R3"] = 1.3 * Rc
            self._initial_params["a1"] = 0.3 * amax
            self._initial_params["a2"] = amax
            self._initial_params["a3"] = 0.3 * amax
            self._initial_params["sigma1"] = 0.5 * sigma
            self._initial_params["sigma2"] = sigma
            self._initial_params["sigma3"] = 0.5 * sigma

        elif form in ['double_powerlaw', 'double_powerlaw_erf']:
            self._initial_params["Rc"] = Rc
            self._initial_params["a"] = amax
            self._initial_params["gamma"] = 1.0
            self._initial_params["alpha1"] = 3.0
            self._initial_params["alpha2"] = -3.0

            if form == 'double_powerlaw':
                self._initial_params["R1"] = None
                self._initial_params["R2"] = None
                self._initial_params["l1"] = None
                self._initial_params["l2"] = None
            else:
                self._initial_params["R1"] = 0.5 * Rc
                self._initial_params["R2"] = 1.5 * Rc
                self._initial_params["l1"] = Rc
                self._initial_params["l2"] = Rc

        elif form == 'double_powerlaw_gauss':
            self._initial_params["a1"] = amax
            self._initial_params["a2"] = 0.5 * amax
            self._initial_params["R1"] = Rc
            self._initial_params["R2"] = Rc + 2.0
            self._initial_params["alpha1"] = 3.0
            self._initial_params["alpha2"] = -3.0          
            self._initial_params["sigma"] = sigma
            self._initial_params["gamma"] = 1.0

        elif form == 'double_powerlaw_double_gauss':
            self._initial_params["a1"] = -0.3 * amax
            self._initial_params["a2"] = -0.3 * amax
            self._initial_params["a3"] = amax
            self._initial_params["R1"] = 0.7 * Rc
            self._initial_params["R2"] = 1.3 * Rc
            self._initial_params["R3"] = Rc
            self._initial_params["alpha1"] = 3.0
            self._initial_params["alpha2"] = -3.0          
            self._initial_params["sigma1"] = 0.1 * sigma
            self._initial_params["sigma2"] = 0.1 * sigma
            self._initial_params["gamma"] = 1.0               

        elif form == 'single_erf_powerlaw':
            self._initial_params["a"] = amax
            self._initial_params["Rc"] = Rc
            self._initial_params["alpha"] = 2.0
            self._initial_params["sigma"] = 0.1 * sigma

        elif form == 'double_erf_powerlaw':
            self._initial_params["a"] = amax
            self._initial_params["R1"] = 0.5 * Rc
            self._initial_params["R2"] = 1.5 * Rc
            self._initial_params["sigma1"] = 0.1 * sigma
            self._initial_params["sigma2"] = 0.1 * sigma
            self._initial_params["alpha"] = 1.0

        else:
            raise ValueError(f"{form} invalid")
        
        optimizer = optax.adam(self._learn_rate)

        return self._fit(self._initial_params, optimizer, self._niter)


    @property
    def functional_form(self):
        """String of the name of the fit functional form"""
        return self._form
    
    @property
    def initial_params(self):
        """Dictionary of initial guess values for model parameters"""
        return self._initial_params
    
    @property    
    def bestfit_params(self):
        """Dictionary of best fit values for model parameters"""
        return self._params
    
    @property
    def loss_history(self):
        """Array of loss values over optimization loop"""
        return self._loss_history


"""This module contains functions defining parametric forms for radial brightness profiles 
(written by Jeff Jennings, parametric forms supplied by Brianna Zawadzki)."""

import jax.numpy as jnp
from jax.scipy.special import erf
import optax

def gauss(r: jnp.ndarray, Rc: jnp.float32, a: jnp.float32, sigma: jnp.float32):
    """
    Gaussian function \Sigma(r) of the form: 

        ..math::

            \Sigma(r) = a * \exp(-(r - R_c)^2 / (2 * \sigma^2)
    """        

    return a * jnp.exp(-(r - Rc) ** 2 / (2 * sigma ** 2))


def asym_gauss(params: optax.Params, r: jnp.ndarray):
    """
    Assymetric (piecewise) Gaussian function \Sigma(r) of the form:

        .. math::

            \Sigma(r) = a * \exp(-(r - Rc)^2 / (2 * \sigma_1^2) \mathrm{for} r < R_c, 
              \Sigma(r) = a * \exp(-(r - Rc)^2 / (2 * \sigma_2^2) \mathrm{for} r \geq R_c
    """      

    return jnp.piecewise(r, [r < params['Rc'], r >= params['Rc']], 
                        [lambda r : gauss(r, params['Rc'], params['a'], params['sigma1']),
                         lambda r : gauss(r, params['Rc'], params['a'], params['sigma2'])]
                         )


def triple_gauss(params: optax.Params, r: jnp.ndarray):
    """
    Sum of three Gaussian functions \Sigma(r)
    """      

    gauss1 = gauss(r, params['R1'], params['a1'], params['sigma1'])
    gauss2 = gauss(r, params['R2'], params['a2'], params['sigma2'])
    gauss3 = gauss(r, params['R3'], params['a3'], params['sigma3'])

    return gauss1 + gauss2 + gauss3


def double_powerlaw(r: jnp.ndarray, Rc: jnp.float32, a: jnp.ndarray, alpha1: jnp.float32, 
                    alpha2: jnp.float32, gamma: jnp.float32):
    """
    Double power law function \Sigma(r) of the form:

        .. math::

            \Sigma(r) = a * [(r / R_c)^{-\alpha_1 * \gamma} + (r / R_c)^{-\alpha_2 * \gamma}]^{-1 / \gamma}
    """

    return a * ((r / Rc) ** (-alpha1 * gamma) + (r / Rc) ** (-alpha2 * gamma)) ** (-1 / gamma)

    
def double_powerlaw_erf(params: optax.Params, r: jnp.ndarray):
    """
    Double power law function \Sigma(r), optionally with limits R1 and R2 that
    multiply the double power law by error functions of the form:

        .. math::

            \Sigma(r) = a * [(r / R_c)^{\alpha_1 * \gamma} + (r / R_c)^{\alpha_2 * \gamma}]^{-1 / \gamma} * 
              [1 + \erf{(r - R_1) / l_1}] * 
              [1 + \erf{(R_2 - r) / l_2}]
    """      

    dpl = double_powerlaw(r, params['Rc'], params['a'], params['alpha1'], params['alpha2'], params['gamma'])

    if params['R1'] is not None:
        dpl *= (1 + erf((r - params['R1']) / params['l1']))
    if params['R2'] is not None:
        dpl *= (1 + erf((params['R2'] - r) / params['l2']))

    return dpl


def double_powerlaw_gauss(params: optax.Params, r: jnp.ndarray):
    """
    Double power law function with inner Gaussian, \Sigma(r) of the form:

        .. math::

            \Sigma(r) =  a_1 * \exp(-(r - R_1)^2 / (2 * \sigma^2) + 
              a_2 * [(r / R_2)^{\alpha_1 * \gamma} + (r / R_2)^{\alpha_2 * \gamma}]^{-1 / \gamma}
    """      

    gaussian = gauss(r, params['R1'], params['a1'], params['sigma'])

    dpl = double_powerlaw(r, params['R2'], params['a2'], params['alpha1'], params['alpha2'], params['gamma'])
    
    return gaussian * dpl


def double_powerlaw_double_gauss(params: optax.Params, r: jnp.ndarray):
    """
    Double power law function with two Gaussians, \Sigma(r) of the form:

        .. math::

            \Sigma(r) =  a_1 * \exp(-(r - R_1)^2 / (2 * \sigma_1^2) + 
              a_2 * \exp(-(r - R_2)^2 / (2 * \sigma_2^2) + 
              a_3 * [(r / R_3)^{\alpha_1 * \gamma} + (r / R_3)^{\alpha_2 * \gamma}]^{-1 / \gamma}
    """      

    gaussian1 = gauss(r, params['R1'], params['a1'], params['sigma1'])
    gaussian2 = gauss(r, params['R2'], params['a2'], params['sigma2'])

    dpl = double_powerlaw(r, params['R3'], params['a3'], params['alpha1'], params['alpha2'], params['gamma'])
    
    return dpl + gaussian1 + gaussian2


def single_erf_powerlaw(params: optax.Params, r: jnp.ndarray):
    """
    Single error function and power law function \Sigma(r) of the form:

        ..math::

            \Sigma(r) = a * [1 - \erf{(R_c - r) / (\sigma * R_c * \sqrt{2})}] * 
              (r / R_c)^{-\alpha} 
    """      

    inner_edge = 1 - erf((params['Rc'] - r) / \
                         (2 ** 0.5 * params['sigma'] * params['Rc']))

    return params['a'] * inner_edge * (r / params['Rc']) ** -params['alpha']


def double_erf_powerlaw(params: optax.Params, r: jnp.ndarray):
    """
    Double error function and power law function \Sigma(r) of the form:

        ..math::

            \Sigma(r) = [1 - \erf{(R_1 - r) / (\sigma_1 * R_1 * \sqrt{2})}] *
              [1 - \erf{(r - R_2) / (\sigma_2 * R_2 * \sqrt{2})}] *  
              (r / R_1)^{-\alpha}

    """      

    inner_edge = 1 - erf((params['R1'] - r) / \
                         (2 ** 0.5 * params['sigma1'] * params['R1']))
    outer_edge = 1 - erf((r - params['R2']) / \
                         (2 ** 0.5 * params['sigma2'] * params['R2']))

    return params['a'] * inner_edge * outer_edge * (r / params['R1']) ** (-params['alpha'])


"""This module contains functions for plotting pipeline results 
(written by Jeff Jennings)."""

import json 
import numpy as np 
import matplotlib.pyplot as plt 
from matplotlib.gridspec import GridSpec
from matplotlib.colors import Normalize
from scipy.interpolate import interp1d 

from frank.utilities import UVDataBinner, make_image, sweep_profile, generic_dht, jy_convert
from mpol.plot import get_image_cmap_norm

from arksia.input_output import get_vis, load_bestfit_frank_uvtable, load_fits_image, parse_rave_filename
from arksia.extract_radial_profile import radial_profile_from_image
from arksia.imager import dirty_image

def plot_image(image, extent, ax, norm=None, cmap='inferno', title=None, 
                cbar_label=None, cbar_pad=0.1):
    """Plot a 2D image, optionally with colorbar. Args follow `plt.imshow`."""
    if norm is None:
        norm = get_image_cmap_norm(image)

    im = ax.imshow(
        image,
        origin="lower",
        interpolation="none",
        extent=extent,
        cmap=cmap,
        norm=norm
    )    
    
    if cbar_label is not None:
        cbar = plt.colorbar(im, ax=ax, location="left", pad=cbar_pad, shrink=0.7)
        cbar.set_label(cbar_label)
    ax.set_title(title)  


def profile_comparison_figure(fits, model, resid_im_robust=2.0, npix=1200):
    """
    Generate a figure comparing clean, rave, frank radial brightness and visibility profiles.

    Parameters
    ----------
    fits : nested list
        Clean, rave, frank profiles to be plotted. Output of `input_output.load_bestfit_profiles`
    model : dict
        Dictionary containing pipeline parameters
    resid_im_robust : float, default=2.0
        Robust weighting parameter used for imaging frank residual visibilities.
    npix : int, default=1200
        Number of pixels along one axis used to make frank images
        
    Returns
    -------
    fig : `plt.figure` instance
        The generated figure
    """
        
    print('  Figures: making profile comparison figure')

    # load best-fit profiles
    [[rc, Ic, Iec], [grid, Vc]], [[rr, Ir, Ier_lo, Ier_hi], [grid, Vr]], [[rf, If, Ief], [grid, Vf], sol] = fits

    fig = plt.figure(figsize=(10,6))
    fig.suptitle("{} -- robust = {} for clean and rave".format(
        model["base"]["disk"],
        model["clean"]["bestfit"]["robust"])
        )

    gs = GridSpec(4, 2, figure=fig, hspace=0, left=0.09, right=0.97, top=0.94, bottom=0.09)

    ax0 = fig.add_subplot(gs[:3, 0])
    ax3 = fig.add_subplot(gs[3, 0])

    ax1 = fig.add_subplot(gs[:3, 1])
    ax2 = fig.add_subplot(gs[3, 1])

    cols, marks, labs = ['C1', 'C3', 'C2'], ['.', 'x', '+'], ['clean', 'rave', 'frank']
    # clean, rave, frank I(r)
    Is_jy_sr = [Ic, Ir, If]
    # rave fits have dfft lower and upper uncertainties; clean and frank don't
    Ies_jy_sr = [[Iec, Iec], [Ier_lo, Ier_hi], [Ief, Ief]]
    rs = [rc, rr, rf]

    for ii, jj in enumerate(Is_jy_sr):     
        # convert Jy / sterad to mJy / arcsec^2
        I_mjy_as2 = jy_convert(jj, 'sterad_arcsec2') * 1e3
        ax0.plot(rs[ii], I_mjy_as2, c=cols[ii], label=labs[ii])
    
        Ie_lo_mjy_as2 = jy_convert(Ies_jy_sr[ii][0], 'sterad_arcsec2') * 1e3
        Ie_hi_mjy_as2 = jy_convert(Ies_jy_sr[ii][1], 'sterad_arcsec2') * 1e3
        ax0.fill_between(rs[ii], I_mjy_as2 - Ie_lo_mjy_as2, I_mjy_as2 + Ie_hi_mjy_as2, 
                         color=cols[ii], alpha=0.4)

    # plot clean, rave, frank Re(V), binned vis.
    [u, v, vis, weights] = get_vis(model)

    # deproject vis
    up, vp, Vp = sol.geometry.apply_correction(u, v, vis)
    bls = np.sqrt(up**2 + vp**2)

    # plot 1d rave residual brightness   
    rave_resid_im_path = parse_rave_filename(model, file_type='rave_residual_image')
    rave_resid_im = np.load(rave_resid_im_path)
    
    # convert Jy / pixel to Jy / arcsec
    rave_resid_im /= model["rave"]["pixel_scale"] ** 2 

    phis_mod = np.linspace(model["base"]["geom"]["PA"] - 180, 
                                  model["base"]["geom"]["PA"] + 180,
                                  model["clean"]["Nphi"] 
                                  )
    
    rave_resid_r, rave_resid_I = radial_profile_from_image( 
        rave_resid_im, geom=model["base"]["geom"], 
        rmax=max(rr), Nr=len(rr), phis=phis_mod, 
        npix=rave_resid_im.shape[0], pixel_scale=model["rave"]["pixel_scale"],
        bmaj=0, bmin=0, image_rms=0, model_image=True, arcsec2=False
        )

    ax3.plot(rave_resid_r, rave_resid_I * 1e3, c='C3') 

    # load frank residual visibilities (at projected data u,v)
    frank_resid_vis = load_bestfit_frank_uvtable(model, resid_table=True)
    # get pixel scale
    xf, _, _ = make_image(sol, npix, project=True)
    frank_pixel_scale = np.diff(xf).mean()
    # plot 1d frank residual brightness
    frank_resid_im = dirty_image(frank_resid_vis, robust=resid_im_robust, npix=npix, pixel_scale=frank_pixel_scale)


    phis_mod = np.linspace(model["base"]["geom"]["PA"] - 180, 
                                  model["base"]["geom"]["PA"] + 180,
                                  model["clean"]["Nphi"] 
                                  )
    
    frank_resid_r, frank_resid_I = radial_profile_from_image(
        frank_resid_im, geom=model["base"]["geom"], 
        rmax=model["frank"]["rout"], Nr=model["frank"]["N"], phis=phis_mod, 
        npix=model["clean"]["npix"], pixel_scale=model["clean"]["pixel_scale"],
        bmaj=0, bmin=0, image_rms=0, model_image=True, arcsec2=False
        )
           
    ax3.plot(frank_resid_r, frank_resid_I * 1e3, c='C2')

    # bin observed visibilities
    bin_cols, bin_marks = ['k', "#a4a4a4"], ["+", "x"]
    for ii, jj in enumerate(model["plot"]["bin_widths"]):
        bin_vis = UVDataBinner(bls, Vp, weights, jj)
        
        # plot binned Re(V)
        ax1.plot(bin_vis.uv / 1e6, bin_vis.V.real * 1e3, c=bin_cols[ii],
                    marker=bin_marks[ii], ls='None',
                    label=r'Obs., {:.0f} k$\lambda$ bins'.format(jj / 1e3)
                    )
        
    # plot vis fits
    for ii, jj in enumerate([Vc, Vr, Vf]):
        ax1.plot(grid / 1e6, jj * 1e3, c=cols[ii], label=labs[ii])

    # plot clean, rave, frank binned residuals.
    # bin vis fits for residual calculation
    bin_vis = UVDataBinner(bls, Vp, weights, model["plot"]["bin_widths"][-1])    
    _, bin_Vc = generic_dht(rc, Ic, Rmax=sol.Rmax, N=sol._info["N"], 
        grid=bin_vis.uv, inc=0)
    _, bin_Vr = generic_dht(rr, Ir, Rmax=sol.Rmax, N=sol._info["N"], 
        grid=bin_vis.uv, inc=0)
    bin_Vf = sol.predict_deprojected(bin_vis.uv, I=If)

    resid_yscale_guess = []
    for ii, jj in enumerate([bin_Vc, bin_Vr, bin_Vf]):
        # bin vis fit residuals
        resid = bin_vis.V.real - jj
        rmse = np.sqrt(np.mean(resid ** 2))

        resid_yscale_guess.append(resid.mean() + resid.std())

        # plot fit residuals
        ax2.plot(bin_vis.uv / 1e6, resid * 1e3, c=cols[ii], 
                          marker=marks[ii], ls='None',
                        label=r'RMSE {:.3f} mJy'.format(rmse * 1e3))

    # symmetric y-bounds for residuals
    resid_yscale = np.array([resid_yscale_guess]) * 1e3
    ax2.set_ylim(-np.nanmax(resid_yscale), np.nanmax(resid_yscale))

    resid_yscale_I = np.max((np.abs(frank_resid_I).max(), np.abs(rave_resid_I).max())) * 1e3
    ax3.set_ylim(-resid_yscale_I * 1.1, resid_yscale_I * 1.1)

    ax3.set_xlim(ax0.get_xbound())
    ax3.set_ylim()
    ax0.legend()
    ax3.set_xlabel(r'r [arcsec]')
    ax0.set_ylabel(r'I [mJy arcsec$^{-2}$]')
    ax3.set_ylabel(r'Resid. I [mJy arcsec$^{-2}$]')

    ax1.legend(loc=1)
    ax2.legend(fontsize=6, loc=1)
    ax2.set_xlabel(r'Baseline [M$\lambda$]')
    ax1.set_ylabel('Re(V) [mJy]')
    ax2.set_ylabel('Resid. [mJy]')
    
    for ax in [ax1, ax2]:
        ax.set_xlim(0, max(bls) * 1.05 / 1e6)
    ax1.tick_params(labelbottom=False)   

    for ax in [ax0, ax1, ax2, ax3]:
        ax.axhline(0, c='c', ls='--', zorder=10)

    ff = '{}/profile_compare_robust{}.png'.format(
        model["base"]["save_dir"], model["clean"]["robust"])
    print('    saving figure to {}'.format(ff))
    plt.savefig(ff, dpi=300)
    
    return fig


def image_comparison_figure(fits, model, resid_im_robust=2.0, npix=1200):
    """
    Generate a figure comparing clean, rave, frank 2d image

    Parameters
    ----------
    fits : nested list
        Clean, rave, frank profiles to be plotted. Output of `input_output.load_bestfit_profiles`
    model : dict
        Dictionary containing pipeline parameters
    resid_im_robust : float, default=2.0
        Robust weighting parameter used for imaging frank residual visibilities.
    npix : int, default=1200
        Number of pixels along one axis used to make frank images

    Returns
    -------
    fig : `plt.figure` instance
        The generated figure
    """

    print('  Figures: making image comparison figure')
    [[rc, Ic, Iec], [grid, Vc]], [[rr, Ir, Ier_lo, Ier_hi], [grid, Vr]], [[rf, If, Ief], [grid, Vf], sol] = fits

    # get clean images
    base_path = "{}/{}.combined.{}corrected.briggs.{}.{}.{}".format(
        model["base"]["clean_dir"], 
        model["base"]["disk"], 
        model["base"]["SMG_sub"],
        model["clean"]["robust"], 
        model["clean"]["npix"], 
        model["clean"]["pixel_scale"]
        )
    clean_fits = base_path + ".pbcor.fits" 
    model_fits = base_path + ".model.fits"

    clean_image, clean_beam = load_fits_image(clean_fits)
    # convert clean image from Jy / beam to Jy / arcsec^2
    bmaj, bmin = clean_beam * 3600
    print('    clean beam: bmaj {} x bmin {} arcsec'.format(bmaj, bmin))
    beam_area = np.pi * bmaj * bmin / (4 * np.log(2))
    clean_image = clean_image / beam_area 

    # convert clean model image from Jy / pixel to Jy / arcsec^2
    model_image = load_fits_image(model_fits, aux_image=True)    
    model_image = model_image / model["clean"]["pixel_scale"] ** 2

    # set clean image pixel size (assuming square image)
    clean_im_xmax = model["clean"]["pixel_scale"] * model["clean"]["npix"] / 2
    clean_extent = [clean_im_xmax, -clean_im_xmax, -clean_im_xmax, clean_im_xmax]

    # make rave pseudo-2d image
    rave_im_xmax = model["rave"]["pixel_scale"] * len(rr) / 2 
    rave_extent = [rave_im_xmax, -rave_im_xmax, -rave_im_xmax, rave_im_xmax]    
    rave_image, _, _ = sweep_profile(rr, Ir, project=True,
        xmax=rave_im_xmax, ymax=rave_im_xmax, dr=model["rave"]["pixel_scale"], 
        phase_shift=True, geom=sol.geometry
        )
    rave_image = jy_convert(rave_image, 'sterad_arcsec2')

    # make rave residual image (again assuming square images)
    rave_resid_im_path = parse_rave_filename(model, file_type='rave_residual_image')
    rave_resid_im = np.load(rave_resid_im_path)

    # convert Jy / pixel to Jy / arcsec
    rave_resid_im /= model["rave"]["pixel_scale"] ** 2 
    rave_resid_Imax = abs(rave_resid_im).max()

    # make frank pseudo-2d image
    xf, yf, frank_image = make_image(sol, npix, project=True)
    frank_image = frank_image.T
    frank_image = jy_convert(frank_image, 'sterad_arcsec2')
    frank_extent = [xf[-1], xf[0], yf[-1], yf[0]]
    frank_pixel_scale = np.diff(xf).mean()

    frank_resid_vis = load_bestfit_frank_uvtable(model, resid_table=True)
    # generate frank residual image
    frank_resid_im = dirty_image(frank_resid_vis, robust=resid_im_robust, npix=npix, pixel_scale=frank_pixel_scale)
    frank_resid_Imax = abs(frank_resid_im).max()
    frank_resid_extent = [xf[-1], xf[0], yf[0], yf[-1]]

    # make figure
    fig = plt.figure(figsize=(10,6))
    fig.suptitle("{} -- robust = {} for clean, rave; {} for frank imaged residuals.\nclean image colormap forced to minimum of 0.".format(
        model["base"]["disk"],
        model["clean"]["bestfit"]["robust"],
        resid_im_robust)
        )
    gs = GridSpec(2, 3, figure=fig, hspace=0.01, wspace=0.2, left=0.04, right=0.97, top=0.98, bottom=0.01)

    ax4 = fig.add_subplot(gs[0, 0])
    ax5 = fig.add_subplot(gs[1, 0])
    ax6 = fig.add_subplot(gs[0, 2]) 
    ax7 = fig.add_subplot(gs[1, 2])
    ax8 = fig.add_subplot(gs[0, 1])
    ax9 = fig.add_subplot(gs[1, 1])

    # plot clean image
    norm = Normalize(vmin=0, vmax=np.nanmax(clean_image) * 1e3)
    plot_image(clean_image * 1e3, clean_extent, ax4, norm=norm,
               cbar_label='$I_{clean}$ [mJy arcsec$^{-2}$]'
               )

    # plot clean model image
    plot_image(model_image * 1e3, clean_extent, ax5, cmap="Reds",
               norm=get_image_cmap_norm(model_image * 1e3, stretch='asinh'), 
               cbar_label='$I_{clean\ model}$ [mJy arcsec$^{-2}$]'
               )            

    # plot rave pseudo-image
    plot_image(rave_image * 1e3, rave_extent, ax6, 
               cbar_label='$I_{rave}$ [mJy arcsec$^{-2}$]'
               )

    # plot (clean - rave) image using symmetric colormap
    rave_resid_norm = Normalize(vmin=-rave_resid_Imax * 1e3, 
                                 vmax=rave_resid_Imax * 1e3)
    plot_image(rave_resid_im * 1e3, rave_extent, ax7, cmap="RdBu_r",
               norm=rave_resid_norm, 
               cbar_label='$I_{clean - rave}$ [mJy arcsec$^{-2}$]'
               )

    # plot frank pseudo-image 
    plot_image(frank_image * 1e3, frank_extent, ax8,
               cbar_label='$I_{frank}$ [mJy arcsec$^{-2}$]'
               )   

    # plot frank imaged residuals using symmetric colormap
    frank_resid_norm = Normalize(vmin=-frank_resid_Imax * 1e3, 
                                 vmax=frank_resid_Imax * 1e3)
    plot_image(frank_resid_im * 1e3, frank_resid_extent, ax9, cmap="RdBu_r", 
               norm=frank_resid_norm, 
               cbar_label='$\mathcal{F}(V_{frank\ resid.}$) [mJy arcsec$^{-2}$]'
               ) 
    ax9.set_title('Pixel scale {:.1f} mas'.format(frank_pixel_scale * 1e3))

    for ax in [ax4, ax5, ax6, ax7, ax8, ax9]:
        ax.set_xlim(7,-7)
        ax.set_ylim(-7,7)

    ff = '{}/image_compare_robust{}.png'.format(
        model["base"]["save_dir"], model["clean"]["robust"])
    print('    saving figure to {}'.format(ff))
    plt.savefig(ff, dpi=300)
    
    return fig


def frank_image_diag_figure(model, sol, frank_resid_vis, resid_im_robust=2.0, 
                            npix=1800, save_prefix=None
                            ):
    """
    Generate a figure showing frank brightness profile reprojected and swept 
    over 2\pi in azimuth, and reprojected dirty image of frank residual visibilities

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters
    sol : frank _HankelRegressor object
        Reconstructed profile using maximum a posteriori power spectrum
        (see frank.radial_fitters.FrankFitter)     
    frank_resid_vis : list
        frank residual visibilities: u-coordinates, v-coordinates, visibility amplitudes 
        (Re(V) + Im(V) * 1j), weights             
    resid_im_robust : float, default=2.0
        Robust weighting parameter used for imaging frank residual visibilities.
    npix : int, default=2400
        Number of pixels along one axis used to make frank images
    save_prefix : string, default = None
        Prefix for saved figure name. If None, the figure won't be saved

    Returns
    -------
    fig : `plt.figure` instance
        The generated figure
    """
    fig, axes = plt.subplots(1, 2, figsize=(10,5))
    
    # make frank pseudo-2d image
    xf, yf, frank_image = make_image(sol, npix, project=True)
    frank_image = frank_image.T
    frank_image = jy_convert(frank_image, 'sterad_arcsec2')
    frank_extent = [xf[-1], xf[0], yf[-1], yf[0]]
    frank_pixel_scale = np.diff(xf).mean()

    # make frank residual image
    frank_resid_im = dirty_image(frank_resid_vis, robust=resid_im_robust, npix=npix, pixel_scale=frank_pixel_scale)
    frank_resid_Imax = abs(frank_resid_im).max()    
    frank_resid_extent = [xf[-1], xf[0], yf[0], yf[-1]]

    # plot frank pseudo-image 
    plot_image(frank_image * 1e3, frank_extent, axes[0], 
               cbar_label='$I_{frank}$ [mJy arcsec$^{-2}$]'
               )   

    # plot frank imaged residuals using symmetric colormap
    frank_resid_norm = Normalize(vmin=-frank_resid_Imax * 1e3, 
                                 vmax=frank_resid_Imax * 1e3)
    plot_image(frank_resid_im * 1e3, frank_resid_extent, axes[1], cmap="RdBu_r", 
               norm=frank_resid_norm, 
               cbar_label='$\mathcal{F}(V_{frank\ resid.}$) [mJy arcsec$^{-2}$]'
               ) 

    for ax in axes:
        ax.set_xlim(7,-7)
        ax.set_ylim(-7,7)

    fig.suptitle("{} -- robust {} for imaged residuals. Pixel scale {:.1f} mas".format(model["base"]["disk"],
                                                                                   resid_im_robust,
                                                                                   frank_pixel_scale * 1e3)
                                                                                   )
    
    if save_prefix:
        plt.savefig(save_prefix + '_frank_images.png', dpi=300,
                    bbox_inches='tight')
        plt.close()

    return fig, axes


def aspect_ratio_figure(model):
    """
    Generate a figure showing results of vertical inference with frank 1+1D

    Parameters
    ----------
    model : dict
        Dictionary containing pipeline parameters
        
    Returns
    -------
    fig : `plt.figure` instance
        The generated figure
    """
    
    print('  Figures: making aspect ratio figure')

    # log evidence results over grid of alpha, wsmooth, h values
    alphas, wsmooths, hs, logevs = np.genfromtxt("{}/vertical_inference_frank.txt".format(model["base"]["save_dir"])).T

    # size of h grid
    nh = len(np.unique(hs))
    # first idx of each unique (alpha, wsmooth) pair (used to group results by priors)
    idx = np.arange(0, len(logevs), nh)

    fig, axes = plt.subplots(ncols=1, nrows=len(idx), figsize=(7.5,5))
    fig.suptitle(model["base"]["disk"])

    for ii, jj in enumerate(idx):
        alpha, wsmooth = alphas[jj], wsmooths[jj]
        if len(idx) == 1:
            h, logev = hs * 1, logevs * 1
            axes = [axes]
        else:
            h, logev = hs[jj:jj + np.diff(idx)[0]], logevs[jj:jj + np.diff(idx)[0]]
        
        # interpolate
        h_interp = interp1d(h, logev, kind='quadratic')
        hgrid = np.logspace(np.log10(h[0]), np.log10(h[-1]), 300)
        logev_fine = h_interp(hgrid)

        # normalize
        logev -= logev_fine.max()
        logev_fine -= logev_fine.max()
   
        # h is log-spaced 
        dh = hgrid * (hgrid[1] / hgrid[0] - 1)

        # cumulative distribution
        cdf = np.cumsum(10 ** logev_fine * dh)
        cdf /= cdf.max()
        cdf, good_idx = np.unique(cdf, return_index=True) # prevent repeat entries of 1.0

        # cumulative dist percentiles
        pct = interp1d(cdf, hgrid[good_idx], kind='quadratic')
        h16, h50, h84 = pct([0.16, 0.5, 0.84])

        logp_fine = 10 ** logev_fine
        logp = 10 ** logev

        # point estimate of h 
        hmax = hgrid[logp_fine.argmax()]

        axes[ii].plot(h, logp, 'r.', label='PDF samples', zorder=10)
        axes[ii].plot(hgrid, logp_fine, 'k', label='PDF interp.')

        axes[ii].plot(hgrid[good_idx], cdf, 'c', label='CDF')

        axes[ii].axvline(h16, ls='--', c='g', label='$\mu$')
        axes[ii].axvline(h50, ls='--', c='c', label='$-1\sigma$')
        axes[ii].axvline(h84, ls='--', c='r', label='$+1\sigma$')
        axes[ii].axvline(hmax, ls='--', c='m', label='point estimate')

        axes[ii].set_xscale('log')

        axes[ii].set_title(r'$\alpha$ = {}, w$_{{smooth}}$ = {:.0e}, h={:.3f}$_{{-{:.3f}}}^{{+{:.3f}}}$'.format(alpha, wsmooth, h50, h50 - h16, h84 - h50)) 

    # show labels on last panel
    plt.legend()
    plt.xlabel('h = H / r')
    plt.ylabel(r'Normalized P(h|V, $\beta$)')

    ff = '{}/vertical_inference_frank.png'.format(
        model["base"]["save_dir"])
    print('    saving figure to {}'.format(ff))
    plt.savefig(ff, dpi=300)

    return fig


def parametric_fit_figure(fit, reference, model):
    """
    Generate a figure showing results of parametric fits to nonparametric frank
      brightness profiles.

    Parameters
    ----------
    fit: dict
        Dictionary containing initial and final fit parameters for parametric
        model, and loss values

    reference : list
        Reference profile radial points, brightness, 1 sigma uncertainty

    model : dict
        Dictionary containing pipeline parameters
        
    Returns
    -------
    fig : `plt.figure` instance
        The generated figure
    """    

    print('  Figures: making parametric fit comparison figure')
    
    fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(10,7.5))
    axes = axes.ravel()
    fig.delaxes(axes[3])

    fig.suptitle(f"{model['base']['disk']}: {fit.functional_form} fit to frank profile")

    # print formatted bestfit values in fig
    st = 'Best fit values:\n'
    for i,j in fit.bestfit_params.items():
        if j is not None:
            st += f"{i} = {j:.2f}\n"
    st += f"\nSource distance: {model['base']['dist']:.0f} pc"
    fig.text(0.6, 0.15, st)

    # reference profile radial points, brightness, 1 sigma
    rr, II, ss = reference

    fit_initial = fit.parametric_model(fit.initial_params, rr)
    fit_final = fit.parametric_model(fit.bestfit_params, rr)
    resid = II - fit_final

    axes[0].plot(rr, II / 1e6, 'k', label='frank')
    # frank 1 sigma
    axes[0].fill_between(rr, (II - ss) / 1e6, (II + ss) / 1e6, color='k', alpha=0.4)    

    axes[0].plot(rr, fit_initial / 1e6, 'c', label="initial guess")
    axes[0].plot(rr, fit_final / 1e6, 'r', label="best fit")
    
    axes[2].plot(rr, resid / 1e6, '.', ms=2, c='#a4a4a4', 
                 label = f"mean {np.mean(resid) / 1e6:.4f} MJy sr$^{-1}$"
                 )
    axes[2].axhline(y=0, c='c', ls='--')

    axes[1].semilogy(fit.loss_history, 'b', label=f"final loss {fit.loss_history[-1]:.2f}")

    axes[0].legend()
    axes[0].set_ylabel(r'I [MJy sr$^{-1}$]')

    axes[2].set_xlabel(r'r [arcsec]')
    axes[2].set_ylabel(r'Resid. I [MJy sr$^{-1}$]')

    # make y-axis symmetric about 0
    bound = max(abs(np.asarray(axes[2].get_ylim())))
    axes[2].set_ylim(-bound, bound)
    axes[2].legend()

    axes[1].legend()
    axes[1].set_xlabel(r'Iteration')    
    axes[1].set_ylabel(r'Loss')    

    ff = f"{model['base']['parametric_dir']}/parametric_fit_{fit.functional_form}.png"
    print(f"    saving figure to {ff}")
    plt.savefig(ff, dpi=300)

    return fig


"""This module contains functions to obtain a radial brightness profile from a CLEAN image 
(written by Seba Marino and Jeff Jennings)."""

import numpy as np
from scipy.optimize import fsolve
from frank.utilities import jy_convert 

def radial_profile_from_image(image, geom, rmax, Nr, phis, 
                            npix, pixel_scale, bmaj, bmin,
                            image_rms, error_std=False, arcsec2=True, 
                            simple_ellipse=False, pb_image=None, 
                            model_image=False, verbose=0, **kwargs):
    """
    Function to extract radial profile from an image 
    
    image: 2D array
        clean image
    geom : dict
        Source geometry. Keys:
        dRA: float, RA offset in arcsec
        dDec: float, Dec offset in arcsec
        PA: float, PA of the disc in degrees
        inc: float, inclination of the disc in degrees
    rmax: float, maximum radius in arcsec
    Nr: int, number of radial bins
    phis: ndarray, array containing sample of PA's that will be use to calculate average intensity
    image_rms: float, image rms in Jy/beam
    error_std: boolean, whether to use the dispersion to calculate uncertainty or the image rms
    arcsec2: boolean, whether to return profile in units of Jy/arcsec2 or Jy/beam
    simple_ellipse: Whether the arc length is calculated with an ellipse or simple ellipse. 
    pb_image: 2D array
        primary beam image. If None, the primary beam will be considered constant and equal to 1.    
    """
    inc, PA, dRA, dDec = geom["inc"], geom["PA"], geom["dRA"], geom["dDec"]

    if pb_image is None:
        pb_image = np.ones((npix,npix))        
        
    if model_image:
        if verbose > 0:
            print('      model_image = True --> setting image_rms = 0')
        image_rms = 0
        bmaj = 0
        if arcsec2:
            # convert CLEAN .model image from [Jy / pixel] to [Jy / arcsec^2] 
            image = image / pixel_scale ** 2
    else:
        # convert to arcsec
        bmaj = bmaj * 3600
        bmin = bmin * 3600
        beam_area = np.pi * bmaj * bmin / (4 * np.log(2))
        if verbose > 0:
            print("      beam = {:.2f} x {:.2f} arcsec".format(bmaj, bmin))

        if arcsec2:
            if verbose > 0:
                print('converting image to [Jy / arcsec^2]')
            # convert CLEAN image from [Jy / CLEAN beam] to [Jy / arcsec^2]
            image = image / beam_area
            image_rms = image_rms / beam_area
    
    return radial_profile(image, pb_image, geom, rmax, Nr, phis, image_rms, 
                        bmaj, pixel_scale, npix, error_std, arcsec2, 
                        simple_ellipse, model_image)


def radial_profile(image, pb_image, geom, rmax, Nr, phis, image_rms, 
                    bmaj, pixel_scale, npix, error_std=False, arcsec2=True, 
                    simple_ellipse=False, model_image=False, rescale_flux=True, 
                    verbose=0):

    """
    dRA, dDec are RA DEC offsets in arcsec
    PA and inc are PA and inc of the disc in deg
    rmax [arcsec] is the maximum deprojected radius at which to do the azimuthal averaging
    Nr is the number of radial points to calculate
    phis [deg] is an array with uniform spacing that sets the range of PA at which to do the interpolation (0 is north)

    """    
    inc, PA, dRA, dDec = geom["inc"], geom["PA"], geom["dRA"], geom["dDec"] 

    # polar grid
    rs = np.linspace(0, rmax, Nr)
    if PA < 0: 
        PA = PA + 180
    PA_rad = PA * np.pi / 180
    phis_rad = phis * np.pi / 180
    dphi_rad = abs(phis_rad[1] - phis_rad[0])
    Nphi_rad = len(phis_rad)

    ecc = np.sin(inc * np.pi / 180)
    # aspect ratio between major and minor axis (>=1)
    chi = 1 / (np.sqrt(1 - ecc ** 2))
    

    # calculate radial profile 
    Is = np.zeros((Nr, Nphi_rad))
    Is_pb = np.zeros((Nr, Nphi_rad))

    err_count = 0
    err_count_pb = 0 

    for ii in range(Nr):
        for jj in range(Nphi_rad):
            xx, yy = ellipse(dRA, dDec, phis_rad[jj], chi, rs[ii], PA_rad)
            
            ip = -int(xx / pixel_scale) + npix // 2
            jp = int(yy / pixel_scale) + npix // 2
            try:
                Is[ii, jj] = image[jp, ip]
            except IndexError:
                if err_count == 0:
                    print('Warning: radial profile extends beyond image. Padding with nan.')
                err_count += 1
                Is[ii, jj] = np.nan
            if not model_image:
                try:
                    Is_pb[ii, jj] = pb_image[jp, ip]
                except IndexError:
                    if err_count_pb == 0:
                        print('Warning: radial profile extends beyond PB image. Padding with nan.')
                    err_count_pb += 1
                    Is_pb[ii, jj] = np.nan

    if err_count > 0:
        print('  Image padded with {} nan'.format(err_count))
    if err_count_pb > 0:
        print('  PB image padded with {} nan'.format(err_count_pb))

    # radial intensity [Jy/arcsec]
    Ir = np.nanmean(Is, axis=1) 
    if model_image: 
        if arcsec2:
            Ir = jy_convert(Ir, 'arcsec2_sterad')
        return rs, Ir

    Ir_pb = np.zeros(Nr)
    
    for i in range(Nphi_rad):
        Ir_pb = Ir_pb + (image_rms / Is_pb[:,i]) ** 2
    Ir_pb = np.sqrt(Ir_pb / Nphi_rad)

    # number of independent points 
    if simple_ellipse:
        # normalize
        arclength = (Nphi_rad - 1) * dphi_rad * np.sqrt((1 + (1 / chi) ** 2 ) / 2 ) 
    else:
        arclength, _ = arc_length(1, 1 / chi, phis_rad[0] - PA_rad, phis_rad[-1] - PA_rad)

    if verbose > 0:
        print('      arc length = {:.2f} deg'.format(arclength * 180 / np.pi))
    Nindep = rs * arclength / bmaj
    Nindep[Nindep < 1] = 1
    
    if error_std:
        I_err = np.nanstd(Is, axis=1) / np.sqrt(Nindep)
    else:
        I_err = Ir_pb / np.sqrt(Nindep)

    if arcsec2:
        Ir = jy_convert(Ir, 'arcsec2_sterad')        
        I_err = jy_convert(I_err, 'arcsec2_sterad')

    if rescale_flux:
        Ir = Ir * np.cos(inc * np.pi / 180)
        I_err = I_err * np.cos(inc * np.pi / 180)

    return rs, Ir, I_err


def arc_length(a, b, phi1, phi2, Nint=1000000):
    # a is semi-major axis (=1)
    # b is the semi-minor axis (=1/aspect_ratio)
    # returns arc length from phi1 to phi2 which are defined as 0 at the disc PA and grow anticlockwise in the sky

    # works with delta y / delta x instead of dy/dx
   
    # translates phi1 and phi2 to angles between 0 and 2pi
    phi1 = simple_phi(phi1)
    phi2 = simple_phi(phi2)
    # catches error when phi1 and phi2 are the same if one wants to integrate over whole circumpherence
    if phi2 == phi1: 
        phi1 = 0.0
        phi2 = 2 * np.pi

    # Figure out the right ranges of xs to integrate. Is phi=0.0 or phi=180 contained in range?
    phis = separate_intervals(phi1, phi2)
    
    Nph = len(phis)
    Arc_length = 0.0

    for i in range(Nph - 1):
        phi_i = phis[i]
        phi_ii = phis[i + 1]  
 
        x1 = x_phi(phi_i, a, b)
        x2 = x_phi(phi_ii, a, b)
        dx = (x2 - x1) / (Nint - 1)
        xs = np.arange(x1, x2, dx)

        Arc_length += np.sum(Delta_s(xs, a, b))
        
    return Arc_length, phis

    
def separate_intervals(phi1, phi2):
    # Figure out the right ranges of xs to integrate. Does phi=0.0 pr phi=180 is contained in range?
    phis = [phi1]

    if phi1 > phi2: # passes through zero
        # we need to figure out if it passes through 180 first
        if phi1 < np.pi:
            phis.append(np.pi)
        phis.append(0.0)

    if phi2 > np.pi and phis[-1] < np.pi:
        phis.append(np.pi)
    phis.append(phi2)

    return np.array(phis)


def ellipse(x0, y0, phi, chi, a, PA):
    ## phi is a position angle in the plane of the sky

    # x0,y0 ellipse center
    # phi pa at which calculate x,y phi=0 is +y axis
    # chi aspect ratio of ellipse with chi>1
    # a semi-major axis
    # a/chi semi-minor axis
    # PA  pa of ellipse 0 is north and pi/2 is east

    if a==0.:
        return x0*np.ones_like(phi), y0*np.ones_like(phi)
    
    phipp = phi - PA
    xpp = x_phi(np.pi / 2 - phipp, a/chi, a)
    ypp = y_phi(np.pi / 2 - phipp, a/chi, a)

    xp = xpp * np.cos(PA) + ypp * np.sin(PA)
    yp = -xpp * np.sin(PA) + ypp * np.cos(PA)
    
    xc = xp + x0
    yc = yp + y0
    
    return xc, yc


def simple_phi(phi): 
    # returns phi between 0 and 2pi
    if abs(phi) != 2 * np.pi:
        phir = phi % (2 * np.pi)
    else: 
        phir = phi

    if phir < 0:
        phir = 2 * np.pi + phir

    return phir


def x_phi(phi, a, b):
    phi=simple_phi(phi)
    
    if phi <= np.pi / 2 or phi >= 3 / 2 * np.pi:
        sign = 1.0
    else:
        sign = -1.0

    return sign / np.sqrt(np.tan(phi) ** 2 / b ** 2 + 1 / a ** 2)


def y_phi(phi, a, b):    
    phi=simple_phi(phi)
    if phi >= 0 and phi <= np.pi:
        sign = 1.0
    else:
        sign = -1.0
    
    return sign / np.sqrt(1 / b ** 2 + 1 / (a ** 2 * np.tan(phi) ** 2))


def Delta_s(xs, a, b): 
    # delta s over an ellipse
    ys = np.zeros(len(xs))
    # avoid imaginary numbers 
    mask=((xs <= a ) & (xs >= -a)) 

    if xs[-1] < xs[0]:
        ys[mask]= b * np.sqrt(1 - xs[mask] ** 2 / a ** 2)
    else:
        ys[mask]= -b * np.sqrt(1 - xs[mask] ** 2 / a ** 2)

    return np.sqrt((ys[:-1] - ys[1:]) ** 2 + (xs[:-1] - xs[1:]) ** 2)


# We want to calculate the critical angle phi along an ellipse (from its major axis)
# when the radius r(phi) becomes smaller than the semi-major axis (a) by a certain factor f.

# We can find phi by starting from the definition of the ellipse:
#     x2/a2 + y2/b2 =1 
    
# and rewrite x as l(phi)*cos(phi) and y as l(phi)*sin(phi). We can also write b as
# a*cos(inc) (inclined ring). Then the ratio between a and l is 

# a/l(phi)=sqrt(cos(phi)**2+sin(phi)**2/cos(inc)**2)
def stretching(phi, inc):
    """
    phi and inc in rad
    """
    return np.sqrt(np.cos(phi)**2+np.sin(phi)**2/np.cos(inc)**2)

def find_phic(inc, f):
    """
    find critical phi (phic) when the ratio a/l(phi) = f (f>1).
    inc in radians.
    f>1. 
    """
    
    if 1./np.cos(inc)<f:
        print('All phis satisfy condition, using phic=pi/2')
        return np.pi/2. 
              
    # solve equation stretching-f=0
    func= lambda phi: stretching(phi, inc)-f
    phic = fsolve(func, np.pi/4.) # initial guess phic=45deg
    return phic[0]
